import json
import logging

from causalbench.commons.utils import package_module
from causalbench.modules.module import Module
from causalbench.services.requests import save_module, fetch_module


class Run(Module):

    def __init__(self, module_id: int = None):
        super().__init__(module_id, 0, None)

    def validate(self):
        # TODO: To be implemented
        pass

    def fetch(self) -> str:
        return fetch_module(self.type,
                            self.module_id,
                            self.version,
                            'runs',
                            'downloaded_run.zip')

    def save(self, state: dict, public: bool) -> bool:
        if state.context.id is None or state.context.version is None:
            logging.error(f'Cannot publish run as it is generated by an unpublished context')
            return False

        zip_file = package_module(state, self.package_path)
        self.module_id, self.version = save_module(self.type,
                                                   self.module_id,
                                                   self.version,
                                                   public,
                                                   zip_file,
                                                   'runs',
                                                   'run.zip')
        return self.module_id is not None

    def __str__(self):
        string = '=' * 80 + '\n'

        string += f'Task: {self.task}' + '\n'

        for result in self.results:
            string += '-' * 80 + '\n'

            string += f'Dataset: {result.dataset.name}' + '\n'

            string += f'Model: {result.model.name}' + '\n'
            for key, value in result.model.hyperparameters.items():
                string += f'    {key}: {value}' + '\n'

            string += 'Metrics:' + '\n'
            for metric in result.metrics:
                string += f'    {metric.name}: {metric.output["score"]}' + '\n'
                for key, value in metric.hyperparameters.items():
                    string += f'        {key}: {value}' + '\n'

        string += '=' * 80

        return string



import os
from abc import ABC, abstractmethod
from importlib.util import spec_from_file_location, module_from_spec

from bunch_py3 import Bunch

from causalbench.modules.module import Module
from causalbench.services.requests import fetch_module


class AbstractTask(ABC):

    @abstractmethod
    def helpers(self) -> any:
        raise NotImplementedError

    @abstractmethod
    def model_data_inputs(self) -> dict[str, type]:
        raise NotImplementedError

    @abstractmethod
    def metric_data_inputs(self) -> dict[str, type]:
        raise NotImplementedError

    @abstractmethod
    def metric_model_inputs(self) -> dict[str, type]:
        raise NotImplementedError


class Task(Module):

    def __init__(self, module_id: str = None, zip_file: str = None):
        super().__init__(module_id, 0, zip_file)

    def instantiate(self, arguments: Bunch):
        # TODO: Create the structure of the new instance
        pass

    def validate(self):
        # TODO: Perform logical validation of the structure
        pass

    def fetch(self):
        return fetch_module(self.type,
                            self.module_id,
                            self.version,
                            'tasks',
                            'downloaded_task.zip')

    def save(self, state, public: bool) -> bool:
        # TODO: Save the task
        pass

    def load(self) -> AbstractTask:
        # form the proper file path
        file_path = os.path.join(self.package_path, self.path)

        # load the module
        spec = spec_from_file_location('module', file_path)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)

        # create an instance of the task
        class_name = getattr(module, self.class_name)
        task: AbstractTask = class_name()

        return task



import torch
import torch.optim as optim
import torch.nn.functional as F
from torch.autograd import Variable
from model import ADDSTCN
import random
import pandas as pd
import numpy as np
import heapq
import copy
import os
import sys

def preparedata(file, target):
    """Reads data from csv file and transforms it to two PyTorch tensors: dataset x and target time series y that has to be predicted."""
    df_data = pd.read_csv(file)
    df_y = df_data.copy(deep=True)[[target]]
    df_x = df_data.copy(deep=True)
    df_yshift = df_y.copy(deep=True).shift(periods=1, axis=0)
    df_yshift[target]=df_yshift[target].fillna(0.)
    df_x[target] = df_yshift
    data_x = df_x.values.astype('float32').transpose()    
    data_y = df_y.values.astype('float32').transpose()
    data_x = torch.from_numpy(data_x)
    data_y = torch.from_numpy(data_y)

    x, y = Variable(data_x), Variable(data_y)
    return x, y


def train(epoch, traindata, traintarget, modelname, optimizer,log_interval,epochs):
    """Trains model by performing one epoch and returns attention scores and loss."""

    modelname.train()
    x, y = traindata[0:1], traintarget[0:1]
        
    optimizer.zero_grad()
    epochpercentage = (epoch/float(epochs))*100
    output = modelname(x)

    attentionscores = modelname.fs_attention
    
    loss = F.mse_loss(output, y)
    loss.backward()
    optimizer.step()

    if epoch % log_interval ==0 or epoch % epochs == 0 or epoch==1:
        print('Epoch: {:2d} [{:.0f}%] \tLoss: {:.6f}'.format(epoch, epochpercentage, loss))

    return attentionscores.data, loss

def findcauses(target, cuda, epochs, kernel_size, layers, 
               log_interval, lr, optimizername, seed, dilation_c, significance, file):
    """Discovers potential causes of one target time series, validates these potential causes with PIVM and discovers the corresponding time delays"""

    print("\n", "Analysis started for target: ", target)
    torch.manual_seed(seed)
    
    X_train, Y_train = preparedata(file, target)
    X_train = X_train.unsqueeze(0).contiguous()
    Y_train = Y_train.unsqueeze(2).contiguous()

    input_channels = X_train.size()[1]
       
    targetidx = pd.read_csv(file).columns.get_loc(target)
          
    model = ADDSTCN(targetidx, input_channels, layers, kernel_size=kernel_size, cuda=cuda, dilation_c=dilation_c)
    if cuda:
        model.cuda()
        X_train = X_train.cuda()
        Y_train = Y_train.cuda()

    optimizer = getattr(optim, optimizername)(model.parameters(), lr=lr)    
    
    scores, firstloss = train(1, X_train, Y_train, model, optimizer,log_interval,epochs)
    firstloss = firstloss.cpu().data.item()
    for ep in range(2, epochs+1):
        scores, realloss = train(ep, X_train, Y_train, model, optimizer,log_interval,epochs)
    realloss = realloss.cpu().data.item()
    
    s = sorted(scores.view(-1).cpu().detach().numpy(), reverse=True)
    indices = np.argsort(-1 *scores.view(-1).cpu().detach().numpy())
    
    #attention interpretation to find tau: the threshold that distinguishes potential causes from non-causal time series
    if len(s)<=5:
        potentials = []
        for i in indices:
            if scores[i]>1.:
                potentials.append(i)
    else:
        potentials = []
        gaps = []
        for i in range(len(s)-1):
            if s[i]<1.: #tau should be greater or equal to 1, so only consider scores >= 1
                break
            gap = s[i]-s[i+1]
            gaps.append(gap)
        sortgaps = sorted(gaps, reverse=True)
        
        for i in range(0, len(gaps)):
            largestgap = sortgaps[i]
            index = gaps.index(largestgap)
            ind = -1
            if index<((len(s)-1)/2): #gap should be in first half
                if index>0:
                    ind=index #gap should have index > 0, except if second score <1
                    break
        if ind<0:
            ind = 0
                
        potentials = indices[:ind+1].tolist()
    print("Potential causes: ", potentials)
    validated = copy.deepcopy(potentials)
    
    #Apply PIVM (permutes the values) to check if potential cause is true cause
    for idx in potentials:
        random.seed(seed)
        X_test2 = X_train.clone().cpu().numpy()
        random.shuffle(X_test2[:,idx,:][0])
        shuffled = torch.from_numpy(X_test2)
        if cuda:
            shuffled=shuffled.cuda()
        model.eval()
        output = model(shuffled)
        testloss = F.mse_loss(output, Y_train)
        testloss = testloss.cpu().data.item()
        
        diff = firstloss-realloss
        testdiff = firstloss-testloss

        if testdiff>(diff*significance): 
            validated.remove(idx) 
    
 
    weights = []
    
    #Discover time delay between cause and effect by interpreting kernel weights
    for layer in range(layers):
        weight = model.dwn.network[layer].net[0].weight.abs().view(model.dwn.network[layer].net[0].weight.size()[0], model.dwn.network[layer].net[0].weight.size()[2])
        weights.append(weight)

    causeswithdelay = dict()    
    for v in validated: 
        totaldelay=0    
        for k in range(len(weights)):
            w=weights[k]
            row = w[v]
            twolargest = heapq.nlargest(2, row)
            m = twolargest[0]
            m2 = twolargest[1]
            if m > m2:
                index_max = len(row) - 1 - max(range(len(row)), key=row.__getitem__)
            else:
                #take first filter
                index_max=0
            delay = index_max *(dilation_c**k)
            totaldelay+=delay
        if targetidx != v:
            causeswithdelay[(targetidx, v)]=totaldelay
        else:
            causeswithdelay[(targetidx, v)]=totaldelay+1
    print("Validated causes: ", validated)
    
    return validated, causeswithdelay, realloss, scores.view(-1).cpu().detach().numpy().tolist()








import TCDF
import argparse
import torch
import pandas as pd
import numpy as np
import networkx as nx
import pylab
import copy
import matplotlib.pyplot as plt
import os
import sys

# os.chdir(os.path.dirname(sys.argv[0])) #uncomment this line to run in VSCode

def check_positive(value):
    """Checks if argument is positive integer (larger than zero)."""
    ivalue = int(value)
    if ivalue <= 0:
         raise argparse.ArgumentTypeError("%s should be positive" % value)
    return ivalue

def check_zero_or_positive(value):
    """Checks if argument is positive integer (larger than or equal to zero)."""
    ivalue = int(value)
    if ivalue < 0:
         raise argparse.ArgumentTypeError("%s should be positive" % value)
    return ivalue

class StoreDictKeyPair(argparse.Action):
    """Creates dictionary containing datasets as keys and ground truth files as values."""
    def __call__(self, parser, namespace, values, option_string=None):
        my_dict = {}
        for kv in values.split(","):
            k,v = kv.split("=")
            my_dict[k] = v
        setattr(namespace, self.dest, my_dict)

def getextendeddelays(gtfile, columns):
    """Collects the total delay of indirect causal relationships."""
    gtdata = pd.read_csv(gtfile, header=None)

    readgt=dict()
    effects = gtdata[1]
    causes = gtdata[0]
    delays = gtdata[2]
    gtnrrelations = 0
    pairdelays = dict()
    for k in range(len(columns)):
        readgt[k]=[]
    for i in range(len(effects)):
        key=effects[i]
        value=causes[i]
        readgt[key].append(value)
        pairdelays[(key, value)]=delays[i]
        gtnrrelations+=1
    
    g = nx.DiGraph()
    g.add_nodes_from(readgt.keys())
    for e in readgt:
        cs = readgt[e]
        for c in cs:
            g.add_edge(c, e)

    extendedreadgt = copy.deepcopy(readgt)
    
    for c1 in range(len(columns)):
        for c2 in range(len(columns)):
            paths = list(nx.all_simple_paths(g, c1, c2, cutoff=2)) #indirect path max length 3, no cycles
            
            if len(paths)>0:
                for path in paths:
                    for p in path[:-1]:
                        if p not in extendedreadgt[path[-1]]:
                            extendedreadgt[path[-1]].append(p)
                            
    extendedgtdelays = dict()
    for effect in extendedreadgt:
        causes = extendedreadgt[effect]
        for cause in causes:
            if (effect, cause) in pairdelays:
                delay = pairdelays[(effect, cause)]
                extendedgtdelays[(effect, cause)]=[delay]
            else:
                #find extended delay
                paths = list(nx.all_simple_paths(g, cause, effect, cutoff=2)) #indirect path max length 3, no cycles
                extendedgtdelays[(effect, cause)]=[]
                for p in paths:
                    delay=0
                    for i in range(len(p)-1):
                        delay+=pairdelays[(p[i+1], p[i])]
                    extendedgtdelays[(effect, cause)].append(delay)

    return extendedgtdelays, readgt, extendedreadgt

def evaluate(gtfile, validatedcauses, columns):
    """Evaluates the results of TCDF by comparing it to the ground truth graph, and calculating precision, recall and F1-score. F1'-score, precision' and recall' include indirect causal relationships."""
    extendedgtdelays, readgt, extendedreadgt = getextendeddelays(gtfile, columns)
    FP=0
    FPdirect=0
    TPdirect=0
    TP=0
    FN=0
    FPs = []
    FPsdirect = []
    TPsdirect = []
    TPs = []
    FNs = []
    for key in readgt:
        for v in validatedcauses[key]:
            if v not in extendedreadgt[key]:
                FP+=1
                FPs.append((key,v))
            else:
                TP+=1
                TPs.append((key,v))
            if v not in readgt[key]:
                FPdirect+=1
                FPsdirect.append((key,v))
            else:
                TPdirect+=1
                TPsdirect.append((key,v))
        for v in readgt[key]:
            if v not in validatedcauses[key]:
                FN+=1
                FNs.append((key, v))
          
    print("Total False Positives': ", FP)
    print("Total True Positives': ", TP)
    print("Total False Negatives: ", FN)
    print("Total Direct False Positives: ", FPdirect)
    print("Total Direct True Positives: ", TPdirect)
    print("TPs': ", TPs)
    print("FPs': ", FPs)
    print("TPs direct: ", TPsdirect)
    print("FPs direct: ", FPsdirect)
    print("FNs: ", FNs)
    precision = recall = 0.

    if float(TP+FP)>0:
        precision = TP / float(TP+FP)
    print("Precision': ", precision)
    if float(TP + FN)>0:
        recall = TP / float(TP + FN)
    print("Recall': ", recall)
    if (precision + recall) > 0:
        F1 = 2 * (precision * recall) / (precision + recall)
    else:
        F1 = 0.
    print("F1' score: ", F1,"(includes direct and indirect causal relationships)")

    precision = recall = 0.
    if float(TPdirect+FPdirect)>0:
        precision = TPdirect / float(TPdirect+FPdirect)
    print("Precision: ", precision)
    if float(TPdirect + FN)>0:
        recall = TPdirect / float(TPdirect + FN)
    print("Recall: ", recall)
    if (precision + recall) > 0:
        F1direct = 2 * (precision * recall) / (precision + recall)
    else:
        F1direct = 0.
    print("F1 score: ", F1direct,"(includes only direct causal relationships)")
    return FP, TP, FPdirect, TPdirect, FN, FPs, FPsdirect, TPs, TPsdirect, FNs, F1, F1direct

def evaluatedelay(extendedgtdelays, alldelays, TPs, receptivefield):
    """Evaluates the delay discovery of TCDF by comparing the discovered time delays with the ground truth."""
    zeros = 0
    total = 0.
    for i in range(len(TPs)):
        tp=TPs[i]
        discovereddelay = alldelays[tp]
        gtdelays = extendedgtdelays[tp]
        for d in gtdelays:
            if d <= receptivefield:
                total+=1.
                error = d - discovereddelay
                if error == 0:
                    zeros+=1
                
            else:
                next
           
    if zeros==0:
        return 0.
    else:
        return zeros/float(total)


def runTCDF(datafile):
    """Loops through all variables in a dataset and return the discovered causes, time delays, losses, attention scores and variable names."""
    df_data = pd.read_csv(datafile)

    allcauses = dict()
    alldelays = dict()
    allreallosses=dict()
    allscores=dict()

    columns = list(df_data)
    for c in columns:
        idx = df_data.columns.get_loc(c)
        causes, causeswithdelay, realloss, scores = TCDF.findcauses(c, cuda=cuda, epochs=nrepochs, 
        kernel_size=kernel_size, layers=levels, log_interval=loginterval, 
        lr=learningrate, optimizername=optimizername,
        seed=seed, dilation_c=dilation_c, significance=significance, file=datafile)

        allscores[idx]=scores
        allcauses[idx]=causes
        alldelays.update(causeswithdelay)
        allreallosses[idx]=realloss

    return allcauses, alldelays, allreallosses, allscores, columns

def plotgraph(stringdatafile,alldelays,columns):
    """Plots a temporal causal graph showing all discovered causal relationships annotated with the time delay between cause and effect."""
    G = nx.DiGraph()
    for c in columns:
        G.add_node(c)
    for pair in alldelays:
        p1,p2 = pair
        nodepair = (columns[p2], columns[p1])

        G.add_edges_from([nodepair],weight=alldelays[pair])
    
    edge_labels=dict([((u,v,),d['weight'])
                    for u,v,d in G.edges(data=True)])
    
    pos=nx.circular_layout(G)
    nx.draw_networkx_edge_labels(G,pos,edge_labels=edge_labels)
    nx.draw(G,pos, node_color = 'white', edge_color='black',node_size=1000,with_labels = True)
    ax = plt.gca()
    ax.collections[0].set_edgecolor("#000000") 

    pylab.show()

def main(datafiles, evaluation):
    if evaluation:
        totalF1direct = [] #contains F1-scores of all datasets
        totalF1 = [] #contains F1'-scores of all datasets

        receptivefield=1
        for l in range(0, levels):
            receptivefield+=(kernel_size-1) * dilation_c**(l)

    for datafile in datafiles.keys(): 
        stringdatafile = str(datafile)
        if '/' in stringdatafile:
            stringdatafile = str(datafile).rsplit('/', 1)[1]
        
        print("\n Dataset: ", stringdatafile)

        # run TCDF
        allcauses, alldelays, allreallosses, allscores, columns = runTCDF(datafile) #results of TCDF containing indices of causes and effects

        print("\n===================Results for", stringdatafile,"==================================")
        for pair in alldelays:
            print(columns[pair[1]], "causes", columns[pair[0]],"with a delay of",alldelays[pair],"time steps.")

        

        if evaluation:
            # evaluate TCDF by comparing discovered causes with ground truth
            print("\n===================Evaluation for", stringdatafile,"===============================")
            FP, TP, FPdirect, TPdirect, FN, FPs, FPsdirect, TPs, TPsdirect, FNs, F1, F1direct = evaluate(datafiles[datafile], allcauses, columns)
            totalF1.append(F1)
            totalF1direct.append(F1direct)

            # evaluate delay discovery
            extendeddelays, readgt, extendedreadgt = getextendeddelays(datafiles[datafile], columns)
            percentagecorrect = evaluatedelay(extendeddelays, alldelays, TPs, receptivefield)*100
            print("Percentage of delays that are correctly discovered: ", percentagecorrect,"%")
            
        print("==================================================================================")
        
        if args.plot:
            plotgraph(stringdatafile, alldelays, columns)

    # In case of multiple datasets, calculate average F1-score over all datasets and standard deviation
    if len(datafiles.keys())>1 and evaluation:  
        print("\nOverall Evaluation: \n")      
        print("F1' scores: ")
        for f in totalF1:
            print(f)
        print("Average F1': ", np.mean(totalF1))
        print("Standard Deviation F1': ", np.std(totalF1),"\n")
        print("F1 scores: ")
        for f in totalF1direct:
            print(f)
        print("Average F1: ", np.mean(totalF1direct))
        print("Standard Deviation F1: ", np.std(totalF1direct))




parser = argparse.ArgumentParser(description='TCDF: Temporal Causal Discovery Framework')

parser.add_argument('--cuda', action="store_true", default=False, help='Use CUDA (GPU) (default: False)')
parser.add_argument('--epochs', type=check_positive, default=1000, help='Number of epochs (default: 1000)')
parser.add_argument('--kernel_size', type=check_positive, default=4, help='Size of kernel, i.e. window size. Maximum delay to be found is kernel size - 1. Recommended to be equal to dilation coeffient (default: 4)')
parser.add_argument('--hidden_layers', type=check_zero_or_positive, default=0, help='Number of hidden layers in the depthwise convolution (default: 0)') 
parser.add_argument('--learning_rate', type=float, default=0.01, help='Learning rate (default: 0.01)')
parser.add_argument('--optimizer', type=str, default='Adam', choices=['Adam', 'RMSprop'], help='Optimizer to use (default: Adam)')
parser.add_argument('--log_interval', type=check_positive, default=500, help='Epoch interval to report loss (default: 500)')
parser.add_argument('--seed', type=check_positive, default=1111, help='Random seed (default: 1111)')
parser.add_argument('--dilation_coefficient', type=check_positive, default=4, help='Dilation coefficient, recommended to be equal to kernel size (default: 4)')
parser.add_argument('--significance', type=float, default=0.8, help="Significance number stating when an increase in loss is significant enough to label a potential cause as true (validated) cause. See paper for more details (default: 0.8)")
parser.add_argument('--plot', action="store_true", default=False, help='Show causal graph (default: False)')
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('--ground_truth',action=StoreDictKeyPair, help='Provide dataset(s) and the ground truth(s) to evaluate the results of TCDF. Argument format: DataFile1=GroundtruthFile1,Key2=Value2,... with a key for each dataset containing multivariate time series (required file format: csv, a column with header for each time series) and a value for the corresponding ground truth (required file format: csv, no header, index of cause in first column, index of effect in second column, time delay between cause and effect in third column)')
group.add_argument('--data', nargs='+', help='(Path to) one or more datasets to analyse by TCDF containing multiple time series. Required file format: csv with a column (incl. header) for each time series')

args = parser.parse_args()

print("Arguments:", args)

if torch.cuda.is_available():
    if not args.cuda:
        print("WARNING: You have a CUDA device, you should probably run with --cuda to speed up training.")
if args.kernel_size != args.dilation_coefficient:
    print("WARNING: The dilation coefficient is not equal to the kernel size. Multiple paths can lead to the same delays. Set kernel_size equal to dilation_c to have exaxtly one path for each delay.")

kernel_size = args.kernel_size
levels = args.hidden_layers+1
nrepochs = args.epochs
learningrate = args.learning_rate
optimizername = args.optimizer
dilation_c = args.dilation_coefficient
loginterval = args.log_interval
seed=args.seed
cuda=args.cuda
significance=args.significance

if args.ground_truth is not None:
    datafiles = args.ground_truth
    main(datafiles, evaluation=True)

else:
    datafiles = dict()
    for dataset in args.data:
        datafiles[dataset]=""
    main(datafiles, evaluation=False)



import platform

import psutil
from bunch_py3 import Bunch
from cpuinfo import cpuinfo

from causalbench.commons.disk import Disks
from causalbench.commons.gpu import GPUs


def hwinfo() -> Bunch:
    response = Bunch()

    # platform information
    response.platform = Bunch()
    response.platform.name = platform.platform()
    response.platform.architecture = platform.architecture()[0]

    # CPU information
    response.cpu = Bunch()
    response.cpu.name = cpuinfo.get_cpu_info()['brand_raw']
    response.cpu.architecture = cpuinfo.get_cpu_info()['arch']

    # GPU information
    gpus = GPUs()
    response.gpu = Bunch()
    for gpu in gpus.devices:
        response.gpu[gpu.bus] = Bunch()
        response.gpu[gpu.bus].name = gpu.name
        response.gpu[gpu.bus].driver = gpu.driver
        response.gpu[gpu.bus].memory_total = gpu.memory_total

    # disk IO information
    disks: Disks = Disks()
    response.disk = Bunch()
    for name, disk in disks.devices.items():
        response.disk[name] = disk

    # memory information
    response.memory_total = psutil.virtual_memory().total

    # storage information
    response.storage_total = psutil.disk_usage('/').total

    return response



from setuptools import find_packages, setup

setup(
    name='causalbench',
    packages=find_packages(include=['causalbench']),
    version='0.1.0',
    description='Spatio Temporal Causal Benchmarking Platform',
    author='Me'
)



import torch as th
from torch import nn
import torch.nn.functional as F
from torch.autograd import Variable
from depthwise import DepthwiseNet
from torch.nn.utils import weight_norm
import numpy as np

class ADDSTCN(nn.Module):
    def __init__(self, target, input_size, num_levels, kernel_size, cuda, dilation_c):
        super(ADDSTCN, self).__init__()

        self.target=target
        self.dwn = DepthwiseNet(self.target, input_size, num_levels, kernel_size=kernel_size, dilation_c=dilation_c)
        self.pointwise = nn.Conv1d(input_size, 1, 1)

        self._attention = th.ones(input_size,1)
        self._attention = Variable(self._attention, requires_grad=False)

        self.fs_attention = th.nn.Parameter(self._attention.data)
        
        if cuda:
            self.dwn = self.dwn.cuda()
            self.pointwise = self.pointwise.cuda()
            self._attention = self._attention.cuda()
                  
    def init_weights(self):
        self.pointwise.weight.data.normal_(0, 0.1)       
        
    def forward(self, x):
        y1=self.dwn(x*F.softmax(self.fs_attention, dim=0))
        y1 = self.pointwise(y1) 
        return y1.transpose(1,2)


import json
import logging
import os
from abc import ABC, abstractmethod
from importlib import resources

import jsonschema
import yaml
from bunch_py3 import bunchify, Bunch
from jsonschema.exceptions import ValidationError

from causalbench.commons.utils import extract_module, extract_module_temporary


class Module(ABC):

    def __init__(self, module_id: any, version: int | None, zip_file: str | None):
        # set the module ID and schema name
        self.module_id = module_id
        self.version = version
        self.type = self.__class__.__name__.lower()

        # load the schema
        self.__load_schema()

        # load the module
        self.__load_module(zip_file)

    def __load_schema(self):
        # load schema
        schema_path = str(resources.files(__package__)
                          .joinpath('schema')
                          .joinpath(self.type + '.yaml'))
        with open(schema_path) as f:
            self.schema = yaml.safe_load(f)

    def __load_module(self, zip_file: str):
        # get package
        self.__get_package(zip_file)

        # package path not defined
        if self.package_path is None:
            self.__dict__.update(Bunch())
            return

        # load configuration
        config_path = os.path.join(self.package_path, 'config.yaml')
        with open(config_path) as f:
            entries = yaml.safe_load(f)
            entries = bunchify(entries)

        # set object structure
        self.__dict__.update(entries)

        # validate object structure
        self.__validate()

    def __get_package(self, zip_file):
        # load directly from zip file
        if zip_file is not None:
            # extract zip to temporary directory
            self.package_path = extract_module_temporary(zip_file)

        # load using module ID and version
        elif self.module_id is not None and self.version is not None:
            self.package_path = extract_module(self.module_id, self.version, self.type, self.fetch())

            # # use cached version if available
            # self.package_path = cached_module(self.module_id, self.version, self.type)
            #
            # # cached version is not available
            # if self.package_path is None:
            #     self.package_path = extract_module(self.module_id, self.version, self.type, self.fetch())

        # nothing to load
        else:
            self.package_path = None

    def publish(self, public: bool = False) -> bool:
        if self.module_id is not None and self.version is not None:
            choice = input(f'Are you sure you want to overwrite existing version of {self.type}? [y/N] ')
            if choice.strip() not in ['y', 'Y']:
                return False

        if public:
            choice = input(f'Are you sure you want to publish {self.type} as public? [y/N] ')
            if choice.strip() not in ['y', 'Y']:
                public = False

        return self.save(self.__getstate__(), public)

    def __validate(self):
        config = json.loads(json.dumps(self.__getstate__(), indent=4))
        try:
            jsonschema.validate(instance=config, schema=self.schema)
            logging.debug('Configuration validated successfully')
        except ValidationError as e:
            logging.error(f'Configuration validation error: {e}')

        try:
            self.validate()
            logging.debug('Logic validated successfully')
        except Exception as e:
            logging.error(f'Logic validation error: {e}')

    def __getstate__(self):
        state = bunchify(self.__dict__)

        if 'module_id' in state:
            del state.module_id

        if 'version' in state:
            del state.version

        if 'schema' in state:
            del state.schema

        if 'package_path' in state:
            del state.package_path

        return state

    @abstractmethod
    def validate(self):
        raise NotImplementedError

    @abstractmethod
    def fetch(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def save(self, state: dict, public: bool) -> bool:
        raise NotImplementedError



import pandas as pd
from bunch_py3 import Bunch


class SpatioTemporalData:

    def __init__(self, data: pd.DataFrame):
        self.data = data
        self._index = Bunch()
        self._index.target = None
        self._index.time = None
        self._index.space = None
        self._index.target = None

    @property
    def time(self):
        return self._index.time
    
    @property
    def target(self):
        return self._index.target
    
    @target.setter
    def target(self, value):
        self.target = value

    @time.setter
    def time(self, value):
        self.time = value

    @property
    def space(self):
        return self._index.space

    @space.setter
    def space(self, value):
        self.space = value

    @property
    def target(self):
        return self._index.target

    @target.setter
    def target(self, value):
        self.target = value

    def update_index(self, data: Bunch):
        if 'index' in data:
            index_dict = {}
            for name, col in data.index.items():
                index_col = data.columns[col]
                if data.headers:
                    index = index_col.header
                else:
                    index = self.data.columns[index_col.position]
                index_dict[name] = index
            self.index = index_dict

    @property
    def index(self) -> dict[str, str]:
        return self._index

    @index.setter
    def index(self, index: dict[str, str]):
        for name, col in index.items():
            if name in self._index:
                self._index[name] = col
            else:
                raise IndexError(f'Invalid index: "{name}"')

    def __copy__(self):
        data_object = SpatioTemporalData(self.data)
        data_object._index.time = self.time
        data_object._index.space = self.space
        data_object._index.target = self.target
        return data_object



import atexit
import logging
import time
from enum import Enum
from threading import Thread

from bunch_py3 import Bunch
import pyopencl as cl

try:
    import pynvml
    try:
        pynvml.nvmlInit()
        atexit.register(lambda: pynvml.nvmlShutdown())
    except Exception as e:
        logging.warning(f'Failed to initialize \'pynvml\' library: {e}')
except Exception as e:
    logging.warning(f'Failed to import \'pynvml\' library: {e}')

try:
    from pyadl import ADLManager, ADLDevice, ADLError
except Exception as e:
    logging.warning(f'Failed to import \'pyadl\' library: {e}')


class Vendor(Enum):
    NVIDIA = 0x10DE
    AMD = 0x1002


class GPU:

    def __init__(self, vendor: Vendor, device: any, cl_device: cl.Device):
        self.vendor = vendor
        self.device = device
        self.cl_device = cl_device

    @property
    def uuid(self):
        if self.vendor == Vendor.NVIDIA:
            return pynvml.nvmlDeviceGetUUID(self.device)

        elif self.vendor == Vendor.AMD:
            return self.device.uuid.decode('utf-8')

    @property
    def bus(self):
        if self.vendor == Vendor.NVIDIA:
            return pynvml.nvmlDeviceGetPciInfo(self.device).bus

        elif self.vendor == Vendor.AMD:
            return self.device.busNumber

    @property
    def name(self):
        if self.vendor == Vendor.NVIDIA:
            return pynvml.nvmlDeviceGetName(self.device)

        elif self.vendor == Vendor.AMD:
            if self.cl_device:
                return f'{self.cl_device.board_name_amd} [{self.cl_device.name}]'
            return self.device.adapterName.decode('utf-8')

    @property
    def memory_used(self):
        if self.vendor == Vendor.NVIDIA:
            return pynvml.nvmlDeviceGetMemoryInfo(self.device).used

        elif self.vendor == Vendor.AMD:
            return None

    @property
    def memory_total(self):
        if self.vendor == Vendor.NVIDIA:
            return pynvml.nvmlDeviceGetMemoryInfo(self.device).total

        elif self.vendor == Vendor.AMD:
            if self.cl_device:
                return self.cl_device.global_mem_size
            return None

    @property
    def utilization(self):
        if self.vendor == Vendor.NVIDIA:
            return pynvml.nvmlDeviceGetUtilizationRates(self.device).gpu

        elif self.vendor == Vendor.AMD:
            try:
                return self.device.getCurrentUsage()
            except ADLError:
                return None

    @property
    def driver(self):
        if self.vendor == Vendor.NVIDIA:
            if self.cl_device:
                return self.cl_device.driver_version
            return pynvml.nvmlSystemGetDriverVersion()

        elif self.vendor == Vendor.AMD:
            if self.cl_device:
                return self.cl_device.driver_version
            return None


class GPUs:

    def __init__(self):
        self._devices = []
        nvidia_cl = dict()
        amd_cl = dict()

        # get devices using opencl
        platforms = cl.get_platforms()
        for platform in platforms:
            devices: list[cl.Device] = platform.get_devices()
            for device in devices:
                # NVIDIA
                if device.vendor_id == Vendor.NVIDIA.value:
                    nvidia_cl[device.pci_bus_id_nv] = device

                # AMD
                elif device.vendor_id == Vendor.AMD.value:
                    amd_cl[device.topology_amd.bus] = device

        # get NVIDIA devices using GPUtil
        try:
            device_count = pynvml.nvmlDeviceGetCount()
            devices: list = [pynvml.nvmlDeviceGetHandleByIndex(i) for i in range(device_count)]
            for index, device in enumerate(devices):
                busNumber = pynvml.nvmlDeviceGetPciInfo(device).bus
                if busNumber in nvidia_cl:
                    self._devices.append(GPU(Vendor.NVIDIA, device, nvidia_cl[busNumber]))
        except:
            pass

        # get AMD devices using pyadl
        try:
            devices: list[ADLDevice] = ADLManager.getInstance().getDevices()
            for index, device in enumerate(devices):
                if device.busNumber in amd_cl:
                    self._devices.append(GPU(Vendor.AMD, device, amd_cl[device.busNumber]))
        except:
            pass

    @property
    def devices(self) -> list[GPU]:
        return self._devices


class GPUsProfiler(Thread):

    def __init__(self, gpus: GPUs = None, delay: int=1):
        super(GPUsProfiler, self).__init__()

        if gpus is None:
            gpus = GPUs()

        self.gpus = gpus
        self.stopped = False
        self.delay = delay

        self.idle = dict()
        self.peak = dict()

    def run(self):
        if self.stopped:
            return

        for gpu in self.gpus.devices:
            self.idle[gpu.bus] = self.peak[gpu.bus] = gpu.utilization

        while not self.stopped:
            for gpu in self.gpus.devices:
                utilization = gpu.utilization

                if utilization is not None:
                    if self.idle[gpu.bus] is not None and utilization < self.idle[gpu.bus]:
                        self.idle[gpu.bus] = utilization
                    if self.peak[gpu.bus] is not None and utilization > self.peak[gpu.bus]:
                        self.peak[gpu.bus] = utilization

            time.sleep(self.delay)

    def stop(self):
        self.stopped = True

    @property
    def utilization(self) -> Bunch:
        utilization = Bunch()
        for gpu in self.gpus.devices:
            utilization[gpu.bus] = Bunch()
            utilization[gpu.bus].idle = self.idle[gpu.bus]
            utilization[gpu.bus].peak = self.peak[gpu.bus]
        return utilization



import json
import platform
import plistlib
import subprocess

import psutil
from bunch_py3 import Bunch, bunchify


class Disks:

    def __init__(self):
        system = platform.system()
        if system == 'Windows':
            self._physical_drives: Bunch = self._physical_drives_windows()
        elif system == 'Linux':
            self._physical_drives: Bunch = self._physical_drives_linux()
        elif system == 'Darwin':
            self._physical_drives: Bunch = self._physical_drives_macos()
        else:
            raise NotImplementedError(f'Unsupported operating system: {system}')

    @property
    def devices(self) -> Bunch:
        return self._physical_drives

    @staticmethod
    def _syscall(*args) -> str:
        result = subprocess.run(*args, capture_output=True, text=True)
        return result.stdout

    def _physical_drives_windows(self):
        import wmi
        c = wmi.WMI()

        physical_drives = Bunch()

        for drive in c.Win32_DiskDrive():
            drive_id = drive.DeviceID.strip("\\\\.\\").replace("PHYSICALDRIVE", "PhysicalDrive")

            physical_drives[drive_id] = Bunch()

            physical_drives[drive_id].model = drive.Model.strip()

            physical_drives[drive_id].usage = Bunch()
            physical_drives[drive_id].usage.total = 0
            physical_drives[drive_id].usage.free = 0

            for disk in drive.associators(wmi_result_class='Win32_DiskPartition'):
                for partition in disk.associators(wmi_result_class='Win32_LogicalDisk'):
                    physical_drives[drive_id].usage.total += int(partition.Size)
                    physical_drives[drive_id].usage.free += int(partition.FreeSpace)

            physical_drives[drive_id].usage.used = physical_drives[drive_id].usage.total - physical_drives[
                drive_id].usage.free

            result = self._syscall(['powershell', '-Command', f'Get-PhysicalDisk | Select-Object DeviceID, MediaType | Where-Object {{ $_.DeviceID -eq {drive.Index} }}'])

            if 'SSD' in result:
                physical_drives[drive_id].mediatype = 'SSD'
            elif 'HDD' in result:
                physical_drives[drive_id].mediatype = 'HDD'
            else:
                physical_drives[drive_id].mediatype = 'Unknown'

            physical_drives[drive_id].fusion = None

        return physical_drives

    def _physical_drives_linux(self):
        def recursive_usage(device):
            if 'children' in device:
                for child in device.children:
                    recursive_usage(child)

            elif device.fssize:
                physical_drives[drive_id].usage.total += int(device.fssize)
                physical_drives[drive_id].usage.free += int(device.fsavail)
                physical_drives[drive_id].usage.used += int(device.fsused)

        result = self._syscall(['lsblk', '-S', '-J', '-o', 'NAME,MODEL,ROTA'])
        drives = json.loads(result)
        drives = bunchify(drives)

        physical_drives = Bunch()

        for drive in drives.blockdevices:
            drive_id = drive.name

            physical_drives[drive_id] = Bunch()

            physical_drives[drive_id].model = drive.model.strip()

            result = self._syscall(['lsblk', '-J', '-b', '-o', 'NAME,FSSIZE,FSAVAIL,FSUSED', f'/dev/{drive_id}'])
            device = json.loads(result)
            device = bunchify(device)
            device = device.blockdevices[0]

            physical_drives[drive_id].usage = Bunch()
            physical_drives[drive_id].usage.total = 0
            physical_drives[drive_id].usage.free = 0
            physical_drives[drive_id].usage.used = 0

            recursive_usage(device)

            if not drive.rota:
                physical_drives[drive_id].mediatype = 'SSD'
            else:
                physical_drives[drive_id].mediatype = 'HDD'

            physical_drives[drive_id].fusion = None

        return physical_drives

    def _physical_drives_macos(self):
        result = self._syscall(['diskutil', 'list', '-plist'])
        plist = plistlib.loads(result.encode())

        # map containers to partitions
        partitions = dict()  # partition to container mapping
        fusions = dict()  # partition to container mapping (for fusion drives)

        for disk in plist['AllDisks']:
            result = self._syscall(['diskutil', 'info', '-plist', disk])
            disk_plist = plistlib.loads(result.encode())

            if disk_plist.get('MountPoint'):
                if disk_plist.get('APFSPhysicalStores'):
                    for partition in disk_plist.get('APFSPhysicalStores'):
                        partition = partition['APFSPhysicalStore']

                        if partition not in partitions:
                            partitions[partition] = []
                        partitions[partition].append(disk)

                        if disk_plist.get('Fusion') and partition not in fusions:
                            fusions[partition] = disk_plist.get('APFSContainerReference')

                else:
                    partitions[disk] = []

        # map partitions to drives
        physical_partitions = Bunch()

        for partition, containers in partitions.items():
            result = self._syscall(['diskutil', 'info', '-plist', partition])
            partition_plist = plistlib.loads(result.encode())

            physical_partitions[partition] = Bunch()

            physical_partitions[partition].parent = partition_plist.get('ParentWholeDisk')

            physical_partitions[partition].usage = Bunch()
            physical_partitions[partition].usage.total = partition_plist.get('TotalSize')
            physical_partitions[partition].usage.used = 0

            if partition_plist.get('FreeSpace'):
                physical_partitions[partition].usage.used = physical_partitions[
                                                                partition].usage.total - partition_plist.get(
                    'FreeSpace')

            else:
                for container in containers:
                    result = self._syscall(['diskutil', 'info', '-plist', container])
                    container_plist = plistlib.loads(result.encode())
                    physical_partitions[partition].usage.used += container_plist.get('CapacityInUse')

        # aggregate partition usage for drives
        physical_drives = Bunch()

        for partition_id, partition in physical_partitions.items():
            device_id = partition.parent

            if device_id not in physical_drives.keys():
                physical_drives[device_id] = Bunch()

                physical_drives[device_id].usage = Bunch()
                physical_drives[device_id].usage.total = 0
                physical_drives[device_id].usage.used = 0

            physical_drives[device_id].usage.total += partition.usage.total
            physical_drives[device_id].usage.used += partition.usage.used

        # drive information
        for device_id, physical_drive in physical_drives.items():
            physical_drive.usage.free = physical_drive.usage.total - physical_drive.usage.used

            result = self._syscall(['diskutil', 'info', '-plist', device_id])
            drive_plist = plistlib.loads(result.encode())

            physical_drive.model = drive_plist.get('MediaName').strip()

            if 'SolidState' in drive_plist:
                if drive_plist.get('SolidState'):
                    physical_drive.mediatype = 'SSD'
                else:
                    physical_drive.mediatype = 'HDD'
            else:
                physical_drive.mediatype = 'Unknown'

            if device_id in fusions:
                physical_drives[device_id].fusion = fusions[device_id]
            else:
                physical_drives[device_id].fusion = None

        return physical_drives


class DisksProfiler:

    def __init__(self, disks: Disks = None):
        if disks is None:
            disks = Disks()
        self.disks = disks.devices.keys()
        self._usage = self._get_usage()

    def _get_usage(self) -> Bunch:
        usage: dict = psutil.disk_io_counters(perdisk=True)
        usage: dict = {key: {'read_bytes': value.read_bytes, 'write_bytes': value.write_bytes} for key, value in
                       usage.items() if key in self.disks}
        usage: dict = dict(sorted(usage.items()))
        return bunchify(usage)

    @property
    def usage(self) -> Bunch:
        current_usage = self._get_usage()
        usage: dict = dict()
        for key in self.disks:
            usage[key] = dict()
            usage[key]['read_bytes'] = current_usage[key]['read_bytes'] - self._usage[key]['read_bytes']
            usage[key]['write_bytes'] = current_usage[key]['write_bytes'] - self._usage[key]['write_bytes']
        return bunchify(usage)



import atexit
import logging
import os
import shutil
import tempfile
from pathlib import Path
from zipfile import ZipFile

import yaml
from bunch_py3 import bunchify, Bunch


def parse_arguments(args, keywords):
    # parse the arguments
    if len(args) == 0:
        return bunchify(keywords)
    elif len(args) == 1:
        if isinstance(args[0], Bunch):
            return args[0]
        elif isinstance(args[0], dict):
            return bunchify(args[0])
    else:
        logging.error('Invalid arguments')
        return


def causal_bench_path(*path_list) -> str:
    path: Path = Path.home().joinpath('.causalbench')
    for path_str in path_list:
        path = path.joinpath(str(path_str))
    return str(path)


def cached_module(module_id, version, module_type: str) -> str:
    # form the directory path
    dir_path = causal_bench_path(module_type, module_id, version)

    # check if directory exists
    if os.path.isdir(dir_path):
        return dir_path


def extract_module(module_id, version, module_type: str, zip_file: str) -> str:
    # form the directory path
    dir_path = causal_bench_path(module_type, module_id, version)

    # extract the zip file
    with ZipFile(zip_file, 'r') as zipped:
        zipped.extractall(path=dir_path)

    return dir_path


def extract_module_temporary(zip_file: str) -> str:
    # form the directory path
    dir_path = tempfile.TemporaryDirectory().name
    atexit.register(lambda: shutil.rmtree(dir_path))

    # extract the zip file
    with ZipFile(zip_file, 'r') as zipped:
        zipped.extractall(path=dir_path)

    return dir_path


def package_module(state, package_path: str, entry_point: str = 'config.yaml') -> str:
    zip_file = tempfile.NamedTemporaryFile(delete=True, suffix='.zip').name
    atexit.register(lambda: os.remove(zip_file))

    with ZipFile(zip_file, 'w') as zipped:
        if entry_point:
            zipped.writestr(entry_point, yaml.safe_dump(state, sort_keys=False, indent=4))

        if package_path is not None:
            for root, dirs, files in os.walk(package_path):
                for file in files:
                    file_path = os.path.join(root, file)
                    zipped_file_path = os.path.relpath(os.path.join(root, file), package_path)
                    if zipped_file_path != entry_point:
                        zipped.write(file_path, zipped_file_path)

    return zip_file



import TCDF
import argparse
import torch
import torch.optim as optim
from model import ADDSTCN
import pandas as pd
import numpy as np
import networkx as nx
import pylab
import copy
import matplotlib.pyplot as plt
import os
import sys

# os.chdir(os.path.dirname(sys.argv[0])) #uncomment this line to run in VSCode

def check_positive(value):
    """Checks if argument is positive integer (larger than zero)."""
    ivalue = int(value)
    if ivalue <= 0:
         raise argparse.ArgumentTypeError("%s should be positive" % value)
    return ivalue

def check_zero_or_positive(value):
    """Checks if argument is positive integer (larger than or equal to zero)."""
    ivalue = int(value)
    if ivalue < 0:
         raise argparse.ArgumentTypeError("%s should be positive" % value)
    return ivalue

class StoreDictKeyPair(argparse.Action):
    """Creates dictionary containing datasets as keys and ground truth files as values."""
    def __call__(self, parser, namespace, values, option_string=None):
        my_dict = {}
        for kv in values.split(","):
            k,v = kv.split("=")
            my_dict[k] = v
        setattr(namespace, self.dest, my_dict)

def check_between_zero_and_one(value):
    """Checks if argument is float between zero and 1."""
    fvalue = float(value)
    if fvalue < 0.0 or fvalue > 1.0:
         raise argparse.ArgumentTypeError("%s should be a float between 0 and 1" % value)
    return fvalue

def evaluate_prediction(target, cuda, epochs, kernel_size, layers, 
               loginterval, lr, optimizername, seed, dilation_c, split, file):
    """Runs first part of TCDF to predict one time series and evaluate its accuracy (MASE)."""
    print("\n", "Analysis started for target: ", target)
    torch.manual_seed(seed)
    
    X, Y = TCDF.preparedata(file, target)
    X = X.unsqueeze(0).contiguous()
    Y = Y.unsqueeze(2).contiguous()

    timesteps = X.size()[2]
    if timesteps!=Y.size()[1]:
        print("WARNING: Time series do not have the same length.")
    X_train = X[:,:,:int(split*timesteps)]
    Y_train = Y[:,:int(split*timesteps),:]
    X_test = X[:,:,int(split*timesteps):]
    Y_test = Y[:,int(split*timesteps):,:]

    input_channels = X_train.size()[1]
    targetidx = pd.read_csv(file).columns.get_loc(target)
          
    model = ADDSTCN(targetidx, input_channels, levels, kernel_size=kernel_size, cuda=cuda, dilation_c=dilation_c)
    if cuda:
        model.cuda()
        X_train = X_train.cuda()
        Y_train = Y_train.cuda()
        X_test = X_test.cuda()
        Y_test = Y_test.cuda()

    optimizer = getattr(optim, optimizername)(model.parameters(), lr=lr)    
    
    for ep in range(1, epochs+1):
        scores, realloss = TCDF.train(ep, X_train, Y_train, model, optimizer,loginterval,epochs)
    realloss = realloss.cpu().data.item()

    model.eval()
    output = model(X_test)
    prediction=output.cpu().detach().numpy()[0,:,0]
    T = output.size()[1]
    total_e = 0.
    for t in range(T):
        real = Y_test[:,t,:]
        predicted = output[:,t,:]
        e = abs(real - predicted)
        total_e+=e
    total_e = total_e.cpu().data.item()
    total = 0.
    for t in range(1,T):
        temp = abs(Y_test[:,t,:] - Y_test[:,t-1,:])
        total+=temp
    denom = (T/float(T-1))*total
    denom = denom.cpu().data.item()

    if denom!=0.:
        MASE = total_e/float(denom)
    else:
        MASE = 0.
    
    return MASE, prediction

def plot_predictions(predictions, file):
    """Plots the predicted values of all time series in the dataset"""
    for c in predictions:
        p = predictions[c]
        plt.plot(p,label=c)
        plt.xlabel('Time')
        plt.ylabel('Predicted value')
        plt.title('Dataset %s'%file)
        plt.legend()

    plt.show()

def evaluate(datafile):
    """Collects the predictions of all time series in a dataset and returns overall results."""
    stringdatafile = str(datafile)
    if '/' in stringdatafile:
        stringdatafile = str(datafile).rsplit('/', 1)[1]
    df_data = pd.read_csv(datafile)
    columns = list(df_data)

    MASEs = []
    predictions = dict()
    for c in columns:
        MASE, prediction = evaluate_prediction(c, cuda=cuda, epochs=nrepochs, 
        kernel_size=kernel_size, layers=levels, loginterval=loginterval, 
        lr=learningrate, optimizername=optimizername,
        seed=seed, dilation_c=dilation_c, split=split, file=datafile)
        predictions[c]= prediction
        MASEs.append(MASE)
        allres.append(MASE)
    avg = np.mean(MASEs)
    std = np.std(MASEs)
    
    return allres, avg, std, predictions
parser = argparse.ArgumentParser(description='TCDF: Temporal Causal Discovery Framework')

parser.add_argument('--cuda', action="store_true", default=False, help='Use CUDA (GPU) (default: False)')
parser.add_argument('--epochs', type=check_positive, default=1000, help='Number of epochs (default: 1000)')
parser.add_argument('--kernel_size', type=check_positive, default=4, help='Size of sliding kernel (default: 4)')
parser.add_argument('--hidden_layers', type=check_zero_or_positive, default=0, help='Number of hidden layers in the depthwise convolution (default: 0)') 
parser.add_argument('--learning_rate', type=float, default=0.01, help='Learning rate (default: 0.01)')
parser.add_argument('--optimizer', type=str, default='Adam', choices=['Adam', 'RMSprop'], help='Optimizer to use: Adam or RMSprop (default: Adam)')
parser.add_argument('--log_interval', type=check_positive, default=500, help='Epoch interval to report loss (default: 500)')
parser.add_argument('--seed', type=check_positive, default=1111, help='Random seed (default: 1111)')
parser.add_argument('--dilation_coefficient', type=check_positive, default=4, help='Dilation coefficient, recommended to be equal to kernel size (default: 4)')
parser.add_argument('--plot', action="store_true", default=False, help='Plot predicted time series (default: False)')
parser.add_argument('--train_test_split', type=check_between_zero_and_one, default=0.8, help="Portion of dataset to use for training (default 0.8)")
parser.add_argument('--data', nargs='+', required=True, help='(Path to) Dataset(s) to predict by TCDF containing multiple time series. Required file format: csv with a column (incl. header) for each time series')

args = parser.parse_args()

print("Arguments:", args)

if torch.cuda.is_available():
    if not args.cuda:
        print("WARNING: You have a CUDA device, you should probably run with --cuda to speed up training.")
if args.kernel_size != args.dilation_coefficient:
    print("WARNING: The dilation coefficient is not equal to the kernel size. Multiple paths can lead to the same delays. Set kernel_size equal to dilation_c to have exaxtly one path for each delay.")

kernel_size = args.kernel_size
levels = args.hidden_layers+1
nrepochs = args.epochs
learningrate = args.learning_rate
optimizername = args.optimizer
dilation_c = args.dilation_coefficient
loginterval = args.log_interval
seed=args.seed
cuda=args.cuda
split=args.train_test_split
plot = args.plot
datasets = args.data

evalresults = dict()
allres = []
for datafile in datasets:
    allres,avg,std,predictions = evaluate(datafile)
    evalresults[datafile]=(avg, std)
    print("\nMean Absolute Scaled Error (MASE) averaged over all time series in", datafile,":",evalresults[datafile][0],"with standard deviation",evalresults[datafile][1])
    if plot:
        plot_predictions(predictions,datafile)

if len(datasets)>1:
    overallavg = np.mean(allres)
    overallstd = np.std(allres)
    print("=========================Overall Evaluation====================================")
    print("Average MASE over all datasets: ", overallavg)
    print("Standard Deviation MASE over all datasets: ", overallstd)




import logging
import time

from bunch_py3 import Bunch, bunchify

from causalbench.commons.hwinfo import hwinfo
from causalbench.commons.utils import parse_arguments, package_module, causal_bench_path
from causalbench.modules.scenario import Scenario
from causalbench.modules.dataset import Dataset
from causalbench.modules.metric import Metric
from causalbench.modules.model import Model
from causalbench.modules.module import Module
from causalbench.modules.run import Run
from causalbench.modules.task import Task
from causalbench.services.requests import save_module, fetch_module


class Context(Module):

    def __init__(self, module_id: int = None, version: int = None, zip_file: str = None):
        super().__init__(module_id, version, zip_file)

    def __getstate__(self):
        state = super().__getstate__()

        if 'datasets' in state:
            for dataset in state.datasets:
                if 'object' in dataset:
                    del dataset.object

        if 'models' in state:
            for model in state.models:
                if 'object' in model:
                    del model.object

        if 'metrics' in state:
            for metric in state.metrics:
                if 'object' in metric:
                    del metric.object

        return state

    def validate(self):
        # TODO: To be implemented
        pass

    def fetch(self):
        return fetch_module(self.type,
                            self.module_id,
                            self.version,
                            'pipelines',
                            'downloaded_context.zip')

    def save(self, state, public: bool = False) -> bool:
        for dataset in state.datasets:
            if dataset.id is None:
                logging.error('Cannot publish context as it contains unpublished dataset')
                return False

        for model in state.models:
            if model.id is None:
                logging.error('Cannot publish context as it contains unpublished model')
                return False

        for metric in state.metrics:
            if metric.id is None:
                logging.error('Cannot publish context as it contains unpublished metric')
                return False

        zip_file = package_module(state, self.package_path)
        self.module_id, self.version = save_module(self.type,
                                                   self.module_id,
                                                   self.version,
                                                   public,
                                                   zip_file,
                                                   'pipelines',
                                                   'context.zip')
        return self.module_id is not None

    def execute(self) -> Run | None:
        # execution start time
        start_time = time.time_ns()

        # load the task
        task: Task = Task(module_id=self.task)

        # load the datasets
        datasets = []
        for dataset in self.datasets:
            if 'object' not in dataset:
                dataset.object = Dataset(module_id=dataset.id, version=dataset.version)
            datasets.append((dataset.object, dataset.mapping))

        # load the models
        models = []
        for model in self.models:
            if 'object' not in model:
                model.object = Model(module_id=model.id, version=model.version)
            models.append((model.object, model.hyperparameters))

        # load the metrics
        metrics = []
        for metric in self.metrics:
            if 'object' not in metric:
                metric.object = Metric(module_id=metric.id, version=metric.version)
            metrics.append((metric.object, metric.hyperparameters))

        # create the scenarios
        scenarios = []
        for dataset in datasets:
            for model in models:
                scenario = Scenario(task, dataset, model, metrics)
                scenarios.append(scenario)

        # execute the scenarios
        results = []
        for scenario in scenarios:
            results.append(scenario.execute())

        # execution end time
        end_time = time.time_ns()

        # create run
        run: Run = Run()

        # context
        run.context = Bunch()
        run.context.id = self.module_id
        run.context.version = self.version
        run.context.name = self.name

        # task
        run.task = self.task

        # results
        run.results = results

        # timing
        run.time = Bunch()
        run.time.start = start_time
        run.time.end = end_time
        run.time.duration = end_time - start_time

        # hardware
        run.profiling = hwinfo()

        # convert results to string for human readability
        for result in run.results:
            for output, value in result.model.output.items():
                result.model.output[output] = str(value)

            for metric in result.metrics:
                for output, value in metric.output.items():
                    metric.output[output] = str(value)

        return run

    def __instantiate(self, arguments: Bunch):
        # module ID
        if 'module_id' in arguments:
            self.module_id = arguments.module_id

        # version
        if 'version' in arguments:
            self.version = arguments.version

        # metadata
        self.name = arguments.name
        self.description = arguments.description
        self.task = arguments.task

        # convert datasets to config format
        self.datasets = []
        if isinstance(arguments.datasets, list):
            for dataset in arguments.datasets:
                self_dataset = Bunch()
                self.datasets.append(self_dataset)

                if isinstance(dataset, tuple):
                    # dataset
                    if isinstance(dataset[0], Dataset):
                        self_dataset.id = dataset[0].module_id
                        self_dataset.version = dataset[0].version
                        self_dataset.object = dataset[0]
                    else:
                        raise ValueError('Invalid dataset instance')

                    # data mapping
                    if isinstance(dataset[1], Bunch):
                        self_dataset.mapping = dataset[1]
                    elif isinstance(dataset[1], dict):
                        self_dataset.mapping = bunchify(dataset[1])
                    else:
                        raise ValueError('Invalid data mapping')
                else:
                    raise ValueError('Invalid dataset definition')
        else:
            raise ValueError('Invalid dataset definition')

        # convert models to config format
        self.models = []
        if isinstance(arguments.models, list):
            for model in arguments.models:
                self_model = Bunch()
                self.models.append(self_model)

                if isinstance(model, tuple):
                    # model
                    if isinstance(model[0], Model):
                        self_model.id = model[0].module_id
                        self_model.version = model[0].version
                        self_model.object = model[0]

                        self_model.hyperparameters = Bunch()
                        if hasattr(model[0], 'hyperparameters'):
                            for hyperparameter, defaults in model[0].hyperparameters.items():
                                self_model.hyperparameters[hyperparameter] = defaults.value
                    else:
                        raise ValueError('Invalid model instance')

                    # hyperparameter mapping
                    if isinstance(model[1], dict):
                        for hyperparameter, value in model[1].items():
                            if hyperparameter in self_model.hyperparameters:
                                self_model.hyperparameters[hyperparameter] = value
                            else:
                                raise ValueError(f'Unknown model hyperparameter {hyperparameter}')
                    else:
                        raise ValueError('Invalid model hyperparameter mapping')
                else:
                    raise ValueError('Invalid model definition')
        else:
            raise ValueError('Invalid model definition')

        # convert metrics to config format
        self.metrics = []
        if isinstance(arguments.metrics, list):
            for metric in arguments.metrics:
                self_metric = Bunch()
                self.metrics.append(self_metric)

                if isinstance(metric, tuple):
                    # metric
                    if isinstance(metric[0], Metric):
                        self_metric.id = metric[0].module_id
                        self_metric.version = metric[0].version
                        self_metric.object = metric[0]

                        self_metric.hyperparameters = Bunch()
                        if hasattr(metric[0], 'hyperparameters'):
                            for hyperparameter, defaults in metric[0].hyperparameters.items():
                                self_metric.hyperparameters[hyperparameter] = defaults.value
                    else:
                        raise ValueError('Invalid metric instance')

                    # hyperparameter mapping
                    if isinstance(metric[1], dict):
                        for hyperparameter, value in metric[1].items():
                            if hyperparameter in self_metric.hyperparameters:
                                self_metric.hyperparameters[hyperparameter] = value
                            else:
                                raise ValueError(f'Unknown metric hyperparameter {hyperparameter}')
                    else:
                        raise ValueError('Invalid metric hyperparameter mapping')
                else:
                    raise ValueError('Invalid metric definition')
        else:
            raise ValueError('Invalid metric definition')

        # form the directory path
        self.package_path = causal_bench_path(self.type, self.name)

    @staticmethod
    def create(*args, **keywords):
        # parse arguments
        arguments = parse_arguments(args, keywords)

        # create the instance
        context = Context()
        context.__instantiate(arguments)
        return context



import logging
import os

from bunch_py3 import Bunch

from causalbench.commons import executor
from causalbench.commons.utils import package_module
from causalbench.modules.module import Module
from causalbench.services.requests import fetch_module, save_module


class Metric(Module):

    def __init__(self, module_id: int = None, version: int = None, zip_file: str = None):
        super().__init__(module_id, version, zip_file)

    def validate(self):
        # check if the file exists
        file_path = os.path.join(self.package_path, self.path)
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File '{self.path}' does not exist in package path '{self.package_path}'")

    def fetch(self):
        return fetch_module(self.type,
                            self.module_id,
                            self.version,
                            'metric_version',
                            'downloaded_metric.zip')

    def save(self, state: dict, public: bool = False) -> bool:
        zip_file = package_module(state, self.package_path)
        self.module_id, self.version = save_module(self.type,
                                                   self.module_id,
                                                   self.version,
                                                   public,
                                                   zip_file,
                                                   'metric_version',
                                                   'metric.zip')
        return self.module_id is not None

    def evaluate(self, parameters: Bunch) -> Bunch:
        # form the proper file path
        file_path = os.path.join(self.package_path, self.path)

        # execute the metric
        response: Bunch = executor.execute(file_path, 'evaluate', **parameters)

        logging.info('Executed metric successfully')

        return response



import functools
import os
import platform
import shutil
import tempfile
import time
import tracemalloc
from importlib.metadata import version
from importlib.util import module_from_spec, spec_from_file_location

import pipreqs.pipreqs as pipreqs
from bunch_py3 import Bunch, bunchify

from causalbench.commons.disk import DisksProfiler
from causalbench.commons.gpu import GPUsProfiler


def execute(module_path, function_name, /, *args, **keywords) -> Bunch:
    # load module
    spec = spec_from_file_location('module', module_path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)

    # get function
    func = getattr(module, function_name)

    # define callable function
    newfunc = functools.partial(func, *args, **keywords)

    # create GPU profiler
    gpu_profiler = GPUsProfiler()
    gpu_profiler.start()

    # create disk IO profiler
    disks_profiler = DisksProfiler()

    # get start time
    start_time = time.time_ns()

    # start memory trace
    tracemalloc.start()

    # execute the function
    output = newfunc()

    # get peak traced memory
    _, memory = tracemalloc.get_traced_memory()

    # end memory trace
    tracemalloc.stop()

    # get the end time
    end_time = time.time_ns()

    # get GPU usage
    gpu_profiler.stop()
    gpu_utilization = gpu_profiler.utilization

    # get disk IO information
    disk_io = disks_profiler.usage

    # get python information
    python = platform.python_version()

    # get imports
    imports = get_imports(module_path)

    # form the response
    response = Bunch()

    # output
    if isinstance(output, Bunch):
        response.output = output
    elif isinstance(output, dict):
        response.output = bunchify(output)
    else:
        raise ValueError(f'Unexpected output type: {type(output)}')

    # timing
    response.time = Bunch()
    response.time.start = start_time
    response.time.end = end_time
    response.time.duration = end_time - start_time

    # profiling
    response.profiling = Bunch()
    response.profiling.memory = memory
    response.profiling.gpu = gpu_utilization
    response.profiling.disk = disk_io
    response.profiling.python = python
    response.profiling.imports = imports

    return response


def get_imports(module_path):
    with tempfile.TemporaryDirectory() as temp_dir:
        # set up paths
        file_name = os.path.basename(module_path)
        temp_path = os.path.join(temp_dir, file_name)

        # copy file to a temporary directory
        shutil.copy2(module_path, temp_path)

        # get names of packages
        candidates = pipreqs.get_all_imports(temp_dir)
        candidates = pipreqs.get_pkg_names(candidates)
        if 'causalbench' in candidates:
            candidates.remove('causalbench')

        # get imports using pipreqs
        imports_pipreqs = pipreqs.get_import_local(candidates)
        candidates = [x for x in candidates if
                      # check if candidate is not in exports
                      x.lower() not in [y for x in imports_pipreqs for y in x['exports']]
                      and
                      # check if candidate is not in package names
                      x.lower() not in [x['name'] for x in imports_pipreqs]]
        imports_pipreqs = {x['name']: x['version'] for x in imports_pipreqs}

        # get imports using importlib
        imports_importlib = {candidate: version(candidate) for candidate in candidates}

        # merge the imports
        imports = {**imports_pipreqs, **imports_importlib}

        # sort the imports
        imports = dict(sorted(imports.items()))

        return bunchify(imports)



import logging
import os

import pandas as pd
from bunch_py3 import Bunch

from causalbench.commons.helpers import adjmat_to_graph
from causalbench.commons.utils import package_module
from causalbench.formats import SpatioTemporalData, SpatioTemporalGraph
from causalbench.modules.module import Module
from causalbench.services.requests import save_module, fetch_module


class Dataset(Module):

    def __init__(self, module_id: int = None, version: int = None, zip_file: str = None):
        super().__init__(module_id, version, zip_file)

    def instantiate(self, arguments: Bunch):
        # TODO: Create the structure of the new instance
        pass

    def validate(self):
        # TODO: Perform logical validation of the structure
        pass

    def fetch(self):
        return fetch_module(self.type,
                            self.module_id,
                            self.version,
                            'dataset_version',
                            'downloaded_dataset.zip')

    def save(self, state: dict, public: bool = False) -> bool:
        zip_file = package_module(state, self.package_path)
        self.module_id, self.version = save_module(self.type,
                                                   self.module_id,
                                                   self.version,
                                                   public,
                                                   zip_file,
                                                   'dataset_version',
                                                   'dataset.zip')
        return self.module_id is not None

    def load(self) -> Bunch:
        files = Bunch()

        for file, data in self.files.items():
            # form the proper file path
            file_path = str(os.path.join(self.package_path, data.path))

            # read the file
            file_df = None
            data_object = None
            if data.data == 'dataframe':
                file_df = pd.read_csv(file_path)
                data_object = SpatioTemporalData(file_df)
                data_object.update_index(data)

            elif data.data == 'graph.static':
                file_df = pd.read_csv(file_path, index_col=0)
                data_object = adjmat_to_graph(file_df.to_numpy(), file_df.columns)

            elif data.data == 'graph.temporal':
                file_df = pd.read_csv(file_path)
                data_object = SpatioTemporalGraph(file_df)
                data_object.update_index(data)

            if file_df is None:
                raise ValueError(f'Invalid data type {data.data}')

            # add data object to the dictionary
            files[file] = data_object

            # validate the file structure
            for column, col_data in data.columns.items():
                if data.headers:
                    col_df = file_df[col_data.header]
                else:
                    col_df = file_df[col_data._index]

                if col_data.data == 'integer':
                    if not pd.api.types.is_integer_dtype(col_df):
                        raise TypeError(f'Data type mismatch for column {column}')
                    if 'labels' in col_data:
                        labels = sorted(col_data.labels)
                        data_labels = sorted(file_df[col_data.header].unique())
                        if labels != data_labels:
                            raise ValueError(f'Labels do not match for column {column}')
                    if 'range' in col_data:
                        start = col_data.range.start
                        end = col_data.range.end
                        min1 = min(file_df[col_data.header])
                        max1 = max(file_df[col_data.header])
                        if not (start <= min1 <= end and start <= max1 <= end):
                            raise ValueError(f'Range does not match for column {column}')

                elif col_data.data == 'decimal':
                    if not pd.api.types.is_float_dtype(col_df):
                        raise TypeError(f'Data type mismatch for column {column}')
                    if 'labels' in col_data:
                        labels = sorted(col_data.labels)
                        data_labels = sorted(file_df[col_data.header].unique())
                        if labels != data_labels:
                            raise ValueError(f'Labels do not match for column {column}')
                    if 'range' in col_data:
                        start = col_data.range.start
                        end = col_data.range.end
                        min1 = min(file_df[col_data.header])
                        max1 = max(file_df[col_data.header])
                        if not (start <= min1 <= end and start <= max1 <= end):
                            raise ValueError(f'Range does not match for column {column}')

        logging.info('Loaded dataset successfully')

        return files



import time

from bunch_py3 import Bunch

from causalbench.modules.dataset import Dataset
from causalbench.modules.metric import Metric
from causalbench.modules.model import Model
from causalbench.modules.task import Task


class Scenario:

    def __init__(self,
                 task: Task,
                 dataset: tuple[Dataset, Bunch],
                 model: tuple[Model, Bunch],
                 metrics: list[tuple[Metric, Bunch]]):
        self.task = task
        self.dataset = dataset
        self.model = model
        self.metrics = metrics

    def execute(self) -> Bunch:
        # execution start time
        start_time = time.time_ns()

        # load the task
        task = self.task.load()

        # load data
        data = self.dataset[0].load()

        # # update indices
        # if 'files' in self.dataset[0]:
        #     for file, data_object in data.items():
        #         if file in self.dataset[0].files:
        #             if isinstance(data_object, SpatioTemporalData):
        #                 data_object.update_index(self.dataset[0].files[file])
        #             if isinstance(data_object, SpatioTemporalGraph):
        #                 data_object.update_index(self.dataset[0].files[file])

        # check model compatibility
        if self.model[0].task != self.task.module_id:
            raise TypeError(f'Model "{self.model[0].name}" not compatible with task "{self.task.module_id}"')

        # check metric compatibility
        for metric in self.metrics:
            if metric[0].task != self.task.module_id:
                raise TypeError(f'Metric "{metric[0].name}" not compatible with task "{self.task.module_id}"')

        # map model parameters
        parameters: Bunch = Bunch()
        parameters.update(self.map_parameters(task.model_data_inputs(), data, self.dataset[1]))
        parameters.update(self.model[1])
        parameters.helpers = task.helpers()

        # execute the model
        model_response: Bunch = self.model[0].execute(parameters)

        # metrics
        scores = []
        for self_metric in self.metrics:
            # map metric parameters
            parameters: Bunch = Bunch()
            parameters.update(self.map_parameters(task.metric_data_inputs(), data, self.dataset[1]))
            parameters.update(self.map_parameters(task.metric_model_inputs(), model_response.output))
            parameters.update(self_metric[1])
            parameters.helpers = task.helpers()

            # execute the metric
            metric_response = self_metric[0].evaluate(parameters)
            metric_response.id = self_metric[0].module_id
            metric_response.version = self_metric[0].version
            metric_response.name = self_metric[0].name
            metric_response.hyperparameters = self_metric[1]
            scores.append(metric_response)

        # execution end time
        end_time = time.time_ns()

        # form the response
        response = Bunch()

        # dataset
        response.dataset = Bunch()
        response.dataset.id = self.dataset[0].module_id
        response.dataset.version = self.dataset[0].version
        response.dataset.name = self.dataset[0].name

        # model
        response.model = model_response
        response.model.id = self.model[0].module_id
        response.model.version = self.model[0].version
        response.model.name = self.model[0].name
        response.model.hyperparameters = self.model[1]

        # metrics
        response.metrics = scores

        # timing
        response.time = Bunch()
        response.time.start = start_time
        response.time.end = end_time
        response.time.duration = end_time - start_time

        return response

    @staticmethod
    def map_parameters(fields: dict[str, type], data: Bunch, mapping: Bunch = None) -> Bunch:
        # if no mapping specified, assume the input and output names are same
        if mapping is None:
            mapping = {field: field for field in fields}

        # map data to fields using mapping
        parameters: Bunch = Bunch()

        for field, datatype in fields.items():
            # check if mapping is specified
            if field not in mapping:
                raise ValueError(f'Mapping does not specify a "{field}" field')

            # check if specified mapping exists
            if mapping[field] not in data:
                raise ValueError(f'Parameter "{mapping[field]}" for field "{field}" does not exist')

            # check if datatype of mapped field is correct
            if not isinstance(data[mapping[field]], datatype):
                raise ValueError(f'Parameter "{mapping[field]}" for field "{field}" is not of type {datatype}')

            # perform mapping
            parameters[field] = data[mapping[field]]

        return parameters



import os
import sys
from pathlib import Path

import requests
import yaml

from causalbench.commons.utils import causal_bench_path


def authenticate(config) -> str | None:
    login_url = "https://www.causalbench.org/api/authenticate/login"

    if 'email' in config:
        email = config['email']
    else:
        return

    if 'password' in config:
        password = config['password']
    else:
        return

    # Payload for login request
    payload = {
        'email_id': email,
        'password': password
    }

    try:
        # Sending login request
        response = requests.post(login_url, json=payload)
        response.raise_for_status()  # Raise an exception for HTTP errors
        data = response.json()['data']

        if data is not None:
            return data['access_token']

    except requests.exceptions.RequestException as e:
        print(f'Error occurred: {e}', file=sys.stderr)
        sys.exit(1)


def init_auth():
    # load config from file
    config_path = causal_bench_path('config.yaml')

    # config file does not exist
    if not os.path.isfile(config_path):
        print('Credentials required')
        create_config(config_path)

    # validate config
    while True:
        with open(config_path, 'r') as f:
            config = yaml.safe_load(f)

        # authenticate
        access_token = authenticate(config)

        # authentication successful
        if access_token is not None:
            return access_token

        # authentication failed
        print('Incorrect credentials')
        create_config(config_path)


def create_config(config_path: str):
    Path(config_path).parent.mkdir(parents=True, exist_ok=True)

    email: str = input('email: ')
    password: str = input('password: ')
    print()

    with open(config_path, 'w') as file:
        yaml.safe_dump({'email': email, 'password': password}, file, indent=4)



import numpy as np
import pandas as pd

from causalbench.formats import SpatioTemporalGraph


def adjmat_to_graph(adjmat: np.ndarray, nodes: list[str], weight: str = 'strength') -> SpatioTemporalGraph:
    if weight not in ['strength', 'lag']:
        raise ValueError(f'Invalid type of weight: {weight}')

    data = []

    for index_cause, cause in enumerate(nodes):
        for index_effect, effect in enumerate(nodes):
            if adjmat[index_cause, index_effect] != 0:
                if weight == 'strength':
                    data.append((cause, effect, 0, 0, adjmat[index_cause, index_effect], 0))
                else:
                    data.append((cause, effect, 0, 0, 0, adjmat[index_cause, index_effect]))

    columns = ['cause', 'effect', 'location_cause', 'location_effect', 'strength', 'lag']
    data = pd.DataFrame(data, columns=columns)

    graph = SpatioTemporalGraph(data)
    graph.index = {x: x for x in columns}

    return graph



import pandas as pd
from bunch_py3 import Bunch


class SpatioTemporalGraph:

    def __init__(self, data: pd.DataFrame):
        self.data = data
        self._index = Bunch()
        self._index.cause = None
        self._index.effect = None
        self._index.location_cause = None
        self._index.location_effect = None
        self._index.strength = None
        self._index.lag = None

    @property
    def cause(self):
        return self._index.cause

    @cause.setter
    def cause(self, value):
        self._index.cause = value

    @property
    def effect(self):
        return self._index.effect

    @effect.setter
    def effect(self, value):
        self._index.effect = value

    @property
    def location_cause(self):
        return self._index.location_cause

    @location_cause.setter
    def location_cause(self, value):
        self._index.location_cause = value

    @property
    def location_effect(self):
        return self._index.location_effect

    @location_effect.setter
    def location_effect(self, value):
        self._index.location_effect = value

    @property
    def strength(self):
        return self._index.strength

    @strength.setter
    def strength(self, value):
        self._index.strength = value

    @property
    def lag(self):
        return self._index.lag

    @lag.setter
    def lag(self, value):
        self._index.lag = value

    @property
    def nodes(self) -> list[str]:
        nodes = set()
        nodes.update(self.data[self.cause].tolist())
        nodes.update(self.data[self.effect].tolist())
        return sorted(nodes)

    def update_index(self, data: Bunch):
        if 'index' in data:
            index_dict = {}
            for name, col in data.index.items():
                index_col = data.columns[col]
                if data.headers:
                    index = index_col.header
                else:
                    index = self.data.columns[index_col.position]
                index_dict[name] = index
            self.index = index_dict

    @property
    def index(self) -> dict[str, str]:
        return self._index

    @index.setter
    def index(self, index: dict[str, str]):
        for name, col in index.items():
            if name in self._index:
                self._index[name] = col
            else:
                raise IndexError(f'Invalid index: "{name}"')

    def __copy__(self):
        data_object = SpatioTemporalGraph(self.data)
        data_object._index.cause = self.cause
        data_object._index.effect = self.effect
        data_object._index.location_cause = self.location_cause
        data_object._index.location_effect = self.location_effect
        data_object._index.strength = self.strength
        data_object._index.lag = self.lag
        return data_object



import torch
import torch.nn as nn
from torch.nn.utils import weight_norm
from torch.autograd import Variable

class Chomp1d(nn.Module):
    """PyTorch does not offer native support for causal convolutions, so it is implemented (with some inefficiency) by simply using a standard convolution with zero padding on both sides, and chopping off the end of the sequence."""
    def __init__(self, chomp_size):
        super(Chomp1d, self).__init__()
        self.chomp_size = chomp_size
    def forward(self, x):
        return x[:, :, :-self.chomp_size].contiguous()

class FirstBlock(nn.Module):
    def __init__(self, target, n_inputs, n_outputs, kernel_size, stride, dilation, padding):
        super(FirstBlock, self).__init__()
        
        self.target = target
        self.conv1 = nn.Conv1d(n_inputs, n_outputs, kernel_size,
                                           stride=stride, padding=padding, dilation=dilation, groups=n_outputs)

        self.chomp1 = Chomp1d(padding)
        self.net = nn.Sequential(self.conv1, self.chomp1)      
        self.relu = nn.PReLU(n_inputs)
        self.init_weights()

    def init_weights(self):
        """Initialize weights"""
        self.conv1.weight.data.normal_(0, 0.1) 
        
    def forward(self, x):
        out = self.net(x)
        return self.relu(out)    

class TemporalBlock(nn.Module):
    def __init__(self, n_inputs, n_outputs, kernel_size, stride, dilation, padding):
        super(TemporalBlock, self).__init__()
       
        self.conv1 = nn.Conv1d(n_inputs, n_outputs, kernel_size,
                                           stride=stride, padding=padding, dilation=dilation, groups=n_outputs)
        self.chomp1 = Chomp1d(padding)
        self.net = nn.Sequential(self.conv1, self.chomp1)
        self.relu = nn.PReLU(n_inputs)
        self.init_weights()

    def init_weights(self):
        """Initialize weights"""
        self.conv1.weight.data.normal_(0, 0.1) 
        

    def forward(self, x):
        out = self.net(x)
        return self.relu(out+x) #residual connection

class LastBlock(nn.Module):
    def __init__(self, n_inputs, n_outputs, kernel_size, stride, dilation, padding):
        super(LastBlock, self).__init__()
        
        self.conv1 = nn.Conv1d(n_inputs, n_outputs, kernel_size,
                                           stride=stride, padding=padding, dilation=dilation, groups=n_outputs)
        self.chomp1 = Chomp1d(padding)
        self.net = nn.Sequential(self.conv1, self.chomp1)
        self.linear = nn.Linear(n_inputs, n_inputs)
        self.init_weights()

    def init_weights(self):
        """Initialize weights"""
        self.linear.weight.data.normal_(0, 0.01) 
        
    def forward(self, x):
        out = self.net(x)
        return self.linear(out.transpose(1,2)+x.transpose(1,2)).transpose(1,2) #residual connection

class DepthwiseNet(nn.Module):
    def __init__(self, target, num_inputs, num_levels, kernel_size=2, dilation_c=2):
        super(DepthwiseNet, self).__init__()
        layers = []
        in_channels = num_inputs
        out_channels = num_inputs
        for l in range(num_levels):
            dilation_size = dilation_c ** l
            if l==0:
                layers += [FirstBlock(target, in_channels, out_channels, kernel_size, stride=1, dilation=dilation_size,
                                     padding=(kernel_size-1) * dilation_size)]
            elif l==num_levels-1:
                layers+=[LastBlock(in_channels, out_channels, kernel_size, stride=1, dilation=dilation_size,
                                     padding=(kernel_size-1) * dilation_size)]
            
            else:
                layers += [TemporalBlock(in_channels, out_channels, kernel_size, stride=1, dilation=dilation_size,
                                     padding=(kernel_size-1) * dilation_size)]

        self.network = nn.Sequential(*layers)

    def forward(self, x):
        return self.network(x)



import os.path
import sys
import tempfile

import requests
from requests import JSONDecodeError as RequestsJSONDecodeError
from bunch_py3 import bunchify

from causalbench import access_token


def save_module(module_type, module_id, version, public, input_file, api_base, default_output_file):
    visibility = "public" if public else "private"
    if module_id is None:
        url = f'https://www.causalbench.org/api/{api_base}/upload?visibility={visibility}'
    elif version is None:
        url = f'https://www.causalbench.org/api/{api_base}/upload/{module_id}?visibility={visibility}'
    else:
        url = f'https://www.causalbench.org/api/{api_base}/upload/{module_id}/{version}?visibility={visibility}'

    headers = {
        'Authorization': f'Bearer {access_token}'
    }
    files = {
        'file': (default_output_file, open(input_file, 'rb'), 'application/zip')
    }

    response = requests.post(url, headers=headers, files=files)

    try:
        data = bunchify(response.json())

        if response.status_code == 200:
            if data.version_num == 0:
                print(f'Published {module_type} with module_id={data.id} (visibility={visibility})', file=sys.stderr)
            else:
                print(f'Published {module_type} with module_id={data.id} and version={data.version_num} (visibility={visibility})', file=sys.stderr)
            return data.id, data.version_num

        else:
            if version == 0:
                print(f'Failed to publish {module_type} with module_id={module_id}: {data.msg} ({response.status_code})', file=sys.stderr)
            else:
                print(f'Failed to publish {module_type} with module_id={module_id} and version={version}: {data.msg} ({response.status_code})', file=sys.stderr)
            sys.exit(1)

    except (RequestsJSONDecodeError, AttributeError):
        if version == 0:
            print(f'Failed to publish {module_type} with module_id={module_id}: {response.text} ({response.status_code})', file=sys.stderr)
        else:
            print(f'Failed to publish {module_type} with module_id={module_id} and version={version}: {response.text} ({response.status_code})', file=sys.stderr)
        sys.exit(1)


def fetch_module(module_type, module_id, version, base_api, default_output_file):
    url = f'https://www.causalbench.org/api/{base_api}/download/{module_id}/{version}'
    headers = {
        'Authorization': f'Bearer {access_token}'
    }

    response = requests.get(url, headers=headers)

    if response.status_code == 200:
        # Extract filename from the Content-Disposition header if available
        content_disposition = response.headers.get('Content-Disposition')
        if content_disposition:
            file_name = content_disposition.split('filename=')[-1].strip('"')
        else:
            # Fallback to a default name if the header is not present
            file_name = default_output_file

        file_path = os.path.join(tempfile.gettempdir(), file_name)
        with open(file_path, 'wb') as file:
            file.write(response.content)

        if version == 0:
            print(f'Fetched {module_type} with module_id={module_id}', file=sys.stderr)
        else:
            print(f'Fetched {module_type} with module_id={module_id} and version={version}', file=sys.stderr)
        return file_path

    else:
        try:
            data = bunchify(response.json())

            if version == 0:
                print(f'Failed to fetch {module_type} with module_id={module_id}: {data.msg} ({response.status_code})', file=sys.stderr)
            else:
                print(f'Failed to fetch {module_type} with module_id={module_id} and version={version}: {data.msg} ({response.status_code})', file=sys.stderr)
            sys.exit(1)

        except (RequestsJSONDecodeError, AttributeError):
            if version == 0:
                print(f'Failed to fetch {module_type} with module_id={module_id}: {response.text} ({response.status_code})', file=sys.stderr)
            else:
                print(f'Failed to fetch {module_type} with module_id={module_id} and version={version}: {response.text} ({response.status_code})', file=sys.stderr)
            sys.exit(1)
